unit UImageHint;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Graphics, Types, StdCtrls, ExtCtrls, Forms, StrUtils, Menus,
  JConfig, StrSub, UAnimatedPaintBox, HogeTextView, UTVSub, UViewItem, UDat2HTML, U2chThread,
  U2chBoard, U2chCatList, U2chCat;

type
  TWebBrowserStatusTextChange = procedure(Sender: TObject; const Text: WideString) of object;

  TNestingHint = class;

  TNestPopUpView = class(THogeTextView)
  protected
    TrackMouse:TTrackMouseEvent;
    MouseOn: Boolean;
    FThreadItem: TThreadItem;
    FURL: String;
    Child: TNestPopUpView;
    FStatusText: String;
    LastPt: TPoint;
    ColorDisable: TColor;
    ColorEnable: TColor;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure ON_WM_CREATE(var msg: TMsg); message WM_CREATE;
    procedure PaintWindow(DC: HDC); override;
    procedure MouseLeave(var Msg: TMessage); message WM_MOUSELEAVE;
    procedure ChildMouseLeave(Sender: TObject);
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure StatusTextChange;
    procedure SetThreadItem(Value: TThreadItem);
    function GetEnabled: Boolean; reintroduce;
    procedure SetEnabled(Value: Boolean); reintroduce;
  public
    StartIndex: Integer;
    OnAbort: TNotifyEvent;
    OnMouseEnter: TNotifyEvent;
    OnMouseLeave: TNotifyEvent;
    OnStatusTextChange: TWebBrowserStatusTextChange;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure PopUp(Rect: TRect);
    function CalcAdjstableSize(MaxHeight, MaxWidth: Integer): TRect;
    function Trim: Boolean;
    function DisableIfMouseAway(Pt: TPoint): Boolean;
    property ThreadItem: TThreadItem read FThreadItem write SetThreadItem;
    property StatusText: String read FStatusText;
    property URL: String read FURL;
    property Enabled: Boolean read GetEnabled write SetEnabled;
  end;

  //Bitmap\łqg|bvAbv
  TImageHint = class(THintWindow)
  protected
    FLastURL: String;
  public
    OwnBitmap: TBitmap;
    PaintBox:TAnimatedPaintBox;
    InfoText: TStaticText;
    Info:string;
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
    procedure ActivateHint(Rect: TRect; const AHint: string); override;
    procedure ActivateHintData(Rect: TRect; const AHint: string; AData: Pointer);override;
    procedure ReleaseHandle;
    procedure Paint; override;
  end;


  //HogeTextViewɂ鑽d|bvAbv
  TNestingHint = class(TImageHint)
  protected
    FParent: TWinControl;
    FBaseThread: TThreadItem;
    WorkingD2View: TDat2View;
    procedure SetParent(Value: TWinControl); reintroduce;
    function GetFixed: Boolean;
    procedure SetFixed(Value: Boolean);
    function CreateNestPopUpView: TNestPopUpView;
    function CreateViewOf2chRes(URI: String; Base: TThreadItem; Point: TPoint): TNestPopUpView;
    function CreateTreeView(AThread: TThreadItem; Index: Integer; OutLine: Boolean; Point: TPoint): TNestPopUpView;
    function Decode2chURL(Url: String; var Board: TBoard;
                     var ThreadItem: TThreadItem; var rangearray: TRangeArray): Boolean;
    procedure PopUpViewAt(View: TNestPopUpView; Point: TPoint);
    function PickUpRes(dest: TDatOut; thread: TThreadItem; startLine, endLine: integer): Integer; overload;
    procedure TextViewMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure TextViewStatusTextChange(Sender: TObject; const Text: WideString);
    procedure TextViewMouseLeave(Sender: TObject);
    procedure AbortWork(Sender: TObject);
    procedure AdjustToTextViewLine(var Point: TPoint; Rect: TRect);
  public
    TextView: TNestPopUpView;
    Dat2HTML: TDat2HTML;
    i2ch: TCategoryList;
    ColorOnFix: TColor;
    LinkColor: TColor;
    PopUpMenu: TPopUpMenu;
    destructor Destroy;override;
    procedure ShowTextHint(const str: string; Point: TPoint; MaxWidth: integer = 0; MaxHeight: integer = 0);
    procedure ActivateHint(Rect: TRect; const AHint: string); reintroduce;
    function ActivateHintTextView(ABaseThread: TThreadItem; Url: String): Boolean;
    function ActivateHintTreeView(AThread: TThreadItem; Index: Integer; OutLine: Boolean): Boolean;
    procedure ReleaseHandle;
    function ReleaseHandleIfNotFix: Boolean;
    function CompareURL(ABaseThread: TThreadItem; Url: String): Boolean;
    function Show2chInfo(URI: String; Base: TThreadItem; Point: TPoint): Boolean;
    function HandleAllocated: Boolean; //override;
    function PickUpRes(thread: TThreadItem; startLine, endLine: integer): String; overload;
    function HasTextView: Boolean;
    function CancelFixIfMouseAway(Pt: TPoint): Boolean;
    property Parent: TWinControl read FParent write SetParent;
    property BaseThread: TThreadItem read FBaseThread;
    property Fixed: Boolean read GetFixed write SetFixed;
  end;


implementation

uses
  CommCtrl, Main, UImageViewer;

{ TNestPopUpView }

constructor TNestPopUpView.Create(AOwner: TComponent);
begin
  inherited;
  inherited Enabled := False;
  TrackMouse.cbSize := Sizeof(TTRACKMOUSEEVENT);
  TrackMouse.dwFlags :=TME_LEAVE;
  Width := 20;
  Height := 20;
  BorderWidth := 2;
  Hide;
end;


destructor TNestPopUpView.Destroy;
begin
  if Assigned(FThreadItem) then
    FThreadItem.Release;
  Child.Free;
  inherited;
end;


procedure TNestPopUpView.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.ExStyle := WS_EX_TOOLWINDOW or WS_EX_TOPMOST;
  Params.WindowClass.style := Params.WindowClass.style or CS_SAVEBITS;
end;

//ftHg̘gƃXN[o[`ȂB
procedure TNestPopUpView.ON_WM_CREATE(var msg: TMsg);
begin
  FEditPoint.X := 0;
  FEditPoint.Y := 0;
  FLogicalCaret.X := 0;
  FLogicalCaret.Y := 0;
  FCaretSavedX := 0;
end;

procedure TNestPopUpView.CreateWnd;
begin
  inherited CreateWnd;
  Windows.SetParent(Handle, 0);
  CallWindowProc(DefWndProc, Handle, WM_SETFOCUS, 0, 0);
end;

//THintWindowɎg`
procedure TNestPopUpView.PaintWindow(DC: HDC);
var
  DCE: HDC;
  HBR: HBRUSH;
  R: TRect;
  R2:TRect;
begin
  R := Rect(0, 0, Width, Height);
  R2 := Rect(1, 1, Width - 1, Height - 1);
  DCE := GetWindowDC(Handle);
  try
    HBR := CreateSolidBrush(ColorToRGB(Color));
    ExcludeClipRect(DCE, BorderWidth, BorderWidth, Width - BorderWidth, Height - BorderWidth);
    FillRect(DCE, R2, HBR);
    DrawEdge(DCE, R, BDR_RAISEDOUTER, BF_RECT);
    DeleteObject(HBR);
    inherited PaintWindow(DC);
  finally
    ReleaseDC(Handle, DCE);
  end;
end;


procedure TNestPopUpView.SetThreadItem(Value: TThreadItem);
begin
  if Assigned(FThreadItem) then
    FThreadItem.Release;
  FThreadItem := Value;
  if Assigned(FThreadItem) then
    FThreadItem.AddRef;
end;


procedure TNestPopUpView.ChildMouseLeave(Sender: TObject);
begin
  FreeAndNil(Child);
  if not PtInRect(BoundsRect, Mouse.CursorPos) then
    PostMessage(Self.Handle, WM_MOUSELEAVE, 0, 0);
end;


procedure TNestPopUpView.PopUp(Rect: TRect);
begin
  BoundsRect := Rect;
  Show;
end;


//e߂ĕ\ł鐡@Ԃ
function TNestPopUpView.CalcAdjstableSize(MaxHeight, MaxWidth: Integer): TRect;
var
  i, j: Integer;
  LineBreak, LineWidth, WMax, MarginWidth: Integer;
  attrib, cw: AnsiString;
begin

  LineBreak := FStrings.Count;
  MarginWidth := FLeftMargin + FRightMargin + BorderWidth * 2 + 0;
  WMax := MarginWidth;
  FBitmap.Canvas.Font := Font;

  for i := 0 to FStrings.Count - 1 do begin

    cw := FStrings[i].GetWidthInfo;
    attrib := FStrings[i].FAttrib;
    LineWidth := MarginWidth + FStrings[i].OffsetLeft;

    for j := 1 to Length(cw) do begin
      if Ord(attrib[j]) and htvVMASK = htvVISIBLE then begin
        if Ord(cw[j]) + LineWidth >= MaxWidth then begin
          Inc(LineBreak);
          if LineWidth > WMax then WMax := LineWidth;
          LineWidth := MarginWidth + FStrings[i].OffsetLeft + Ord(cw[j]);
        end else begin
          Inc(LineWidth, Ord(cw[j]));
        end;
      end;
    end;
    if LineWidth > WMax then WMax := LineWidth;
  end;

  Result.Top := 0;
  Result.Left := 0;
  Result.Bottom := BaselineSkip * LineBreak + TopMargin * 2 + BorderWidth * 2;
  if Result.Bottom > MaxHeight then begin
    Result.Bottom := MaxHeight;
    Result.Right := WMax + GetSystemMetrics(SM_CXVSCROLL) + 2;
    if Result.Right > MaxWidth then Result.Right := MaxWidth;
  end else begin
    Result.Right := WMax;
  end;

end;


//̋󔒍s폜
function TNestPopUpView.Trim: Boolean;
var
  i: Integer;
begin
  for i := FStrings.Count - 1 downto 0 do begin
    if Length(FStrings[i].FText) > 0 then
      Break;
    FStrings[i].Free;
    FStrings.Delete(i);
  end;

  if FStrings.Count <= 0 then begin
    Result:= False;
  end else begin
    Refresh;
    Result := True;
  end;
end;

function TNestPopUpView.DisableIfMouseAway(Pt: TPoint): Boolean;
begin

  if not Enabled or PtInRect(BoundsRect, Pt) then begin
    Result := False;
    Exit;
  end;
  if    ((Pt.Y < Top) and (Pt.Y < LastPt.Y))
     or ((Pt.Y > Top + Height) and (Pt.Y > LastPt.Y))
     or ((Pt.X < Left) and (Pt.X < LastPt.X))
     or ((Pt.X > Left + Width) and (Pt.X > LastPt.X)) then begin
    //Lv`
    Enabled := False;
    Result := True;
  end else begin
    Result := False;
  end;
  LastPt := Pt;
end;


//vpeB֌W
function TNestPopUpView.GetEnabled: Boolean;
begin
  Result := inherited Enabled;
end;
//
procedure TNestPopUpView.SetEnabled(Value: Boolean);
begin
  inherited Enabled := Value;
  if Enabled then begin
    Color := ColorEnable;
    LastPt := Mouse.CursorPos;
  end else begin
    Color := ColorDisable;
  end;
end;


//CxgvV[W֌W
procedure TNestPopUpView.MouseLeave(var Msg: TMessage);
var
  CRect: TRect;
begin
  MouseOn := False;

  CRect.Top := BoundsRect.Top + BorderWidth;
  CRect.Bottom := BoundsRect.Bottom - BorderWidth;
  CRect.Left := BoundsRect.Left + BorderWidth;
  CRect.Right := BoundsRect.Right - BorderWidth;

  if PtInRect(CRect, Mouse.CursorPos) then
    Exit;  //XN[o[ŏ̂h

  if not(Assigned(Child)) or not PtInRect(Child.BoundsRect, Mouse.CursorPos) then begin
    if Assigned(OnAbort) then
      OnAbort(Self);
    if Assigned(OnMouseLeave) then
      OnMouseLeave(Self);
  end;
end;
//
procedure TNestPopUpView.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  st:string;
begin
  inherited;
  if not MouseOn then begin
    TrackMouse.hwndTrack := Handle;
    _TrackMouseEvent(@TrackMouse);
    MouseOn := True;
  end;
  if Assigned(Child) then
    Child.DisableIfMouseAway(ClientToScreen(Point(X,Y)));

  st := TVMouseProc(Self, Shift, X, Y);
  if st <> FStatusText then begin
    FStatusText := st;
    StatusTextChange;
  end else if st = '' then begin
    if Assigned(Child) and not Child.Enabled then
      FreeAndNil(Child);
  end;
end;
//
procedure TNestPopUpView.StatusTextChange;
begin
  if Assigned(OnStatusTextChange) then
    OnStatusTextChange(Self, StatusText);
end;


{ TImageHint }
procedure TImageHint.ActivateHint(Rect: TRect; const AHint: string);
begin
  FLastURL := 'Not ThreadData';
  PaintBox.Bitmap:=nil;
  InfoText.Visible := False;
  Canvas.Font.Assign(Font);
  inherited;
end;

procedure TImageHint.ActivateHintData(Rect: TRect; const AHint: string;
  AData: Pointer);
begin
  inherited;
  FLastURL := 'Not ThreadData';
  if AHint <> '' then
    InfoText.Visible := True
  else
    InfoText.Visible := False;
  InfoText.Font := Font;
  InfoText.Caption := AHint;
  if TObject(AData) is TBitmap then
    PaintBox.Bitmap := TBitmap(AData);
end;

procedure TImageHint.ReleaseHandle;
begin
  PaintBox.Bitmap:=nil;//ID:ie1ZUXSN 2xڂ̋Nȍ~G[
  Info:='';

  OwnBitmap.ReleaseHandle;
  inherited;
end;


constructor TImageHint.Create(AOwner: TComponent);
begin
  inherited;
  InfoText := TStaticText.Create(Self);
  InfoText.Parent:=Self;
  InfoText.Align:=alTop;
  InfoText.Visible := False;

  PaintBox:=TAnimatedPaintBox.Create(Self);
  PaintBox.Parent:=Self;
  PaintBox.Align:=alClient;
  PaintBox.Enabled:=False;
  OwnBitmap := TBitmap.Create;
end;

destructor TImageHint.Destroy;
begin
  PaintBox.Free;
  OwnBitmap.Free;
  inherited;
end;

procedure TImageHint.Paint;
var
  R: TRect;
begin
  R := ClientRect;
  Inc(R.Left, 2);
  Inc(R.Top, 2);
  DrawText(Canvas.Handle, PChar(Caption), -1, R, DT_LEFT or DT_NOPREFIX or
    DT_WORDBREAK or DrawTextBiDiModeFlagsReadingOnly);
end;


{ TNestingHint }


destructor TNestingHint.Destroy;
begin
  TextView.Free;
  inherited;
end;


procedure TNestingHint.ActivateHint(Rect: TRect; const AHint: string);
begin
  ReleaseHandle;
  inherited;
end;


function TNestingHint.ActivateHintTextView(ABaseThread: TThreadItem; Url: String): Boolean;
var
  tmpTextView:TNestPopUpView;
  s: String;
begin
  Result := False;
  if Fixed then begin
    Result := True;
    Exit;
  end;
  if ((URL = FLastURL) and (ABaseThread = FBaseThread)) then begin
    if URL <> '' then
      Result := True;
    Exit;
  end;

  ReleaseHandle;
  FLastURL := URL;

  if Url = '' then Exit;
  FBaseThread := ABaseThread;
  if Assigned(FBaseThread) then
    FBaseThread.AddRef;

  if AnsiStartsStr('mailto:', Url) then begin
    s := copy(Url, 8, length(Url));
    if 0 < length(s) then begin
      ShowTextHint(s, Mouse.CursorPos);
      Result := True;
    end;
  end else if Config.hintNestingPopUp then begin
    tmpTextView := CreateViewOf2chRes(Url, FBaseThread, Mouse.CursorPos);
    if Assigned(textview) then //processmessageɂētextviewĂ\
      tmpTextView.Free
    else
      TextView := tmpTextView;
    Result := Assigned(TextView);
    if Result then begin
      TextView.OnMouseLeave := TextViewMouseLeave;
    end else begin
      FLastURL := '';
    end;
  end else begin
    Result := Show2chInfo(Url, FBaseThread, Mouse.CursorPos);
  end;

end;


function TNestingHint.ActivateHintTreeView(AThread: TThreadItem; Index: Integer; OutLine: Boolean): Boolean;
var
  tmpTextView:TNestPopUpView;
begin
  ReleaseHandle;
  FBaseThread := AThread;
  if Assigned(FBaseThread) then
    FBaseThread.AddRef;

  tmpTextView := CreateTreeView(FBaseThread, Index, OutLine, Mouse.CursorPos);
  if Assigned(textview) then //processmessageɂētextviewĂ\
    tmpTextView.Free
  else
    TextView := tmpTextView;
  Result := Assigned(TextView);
  if Result then
    TextView.OnMouseLeave := TextViewMouseLeave;
end;


procedure TNestingHint.ReleaseHandle;
begin
  if Assigned(FBaseThread) then begin
    FBaseThread.Release;
    FBaseThread := nil;
  end;
  if Assigned(WorkingD2View) then begin
    WorkingD2View.Cancel;
    WorkingD2View := nil;
  end;
  FreeandNil(TextView);
  FLastURL := '';
  inherited;
end;


function TNestingHint.ReleaseHandleIfNotFix: Boolean;
begin
  if not Fixed then begin
    ReleaseHandle;
    Result := True;
  end else begin
    Result := False;
  end;
end;

function TNestingHint.HandleAllocated: Boolean;
begin
  Result := inherited HandleAllocated or Assigned(TextView);
end;


procedure TNestingHint.SetParent(Value: TWinControl);
begin
  FParent := Value;
  if Assigned(TextView) then
    TextView.Parent := Value;
end;

function TNestingHint.CancelFixIfMouseAway(Pt: TPoint): Boolean;
begin
  if Assigned(TextView) then
    Result := TextView.DisableIfMouseAway(Pt)
  else
    Result := False;
end;

function TNestingHint.GetFixed: Boolean;
begin
  Result := Assigned(TextView) and TextView.Enabled;
end;

procedure TNestingHint.SetFixed(Value: Boolean);
begin
  if Assigned(TextView) then
    TextView.Enabled := Value;
end;

function TNestingHint.HasTextView;
begin
  Result := Assigned(TextView);
end;



//|bvAbvr[̃}EXI[o[ŐVȃ|bvAbv쐻
procedure TNestingHint.TextViewStatusTextChange(Sender: TObject; const Text: WideString);
var
  View, tmpChild: TNestPopUpView;
  Point: TPoint;
begin
  inherited ReleaseHandle;
  View := TNestPopUpView(Sender);
  if Assigned(View.Child) then begin
    if View.Child.Enabled then begin
      Exit;
    end else begin
      FreeAndNil(View.Child); //Ƃ肠Ô
    end;
  end;
  AbortWork(nil);
  if ImageForm.ShowImageHint(Text, False) then
    Exit;
  Point := Mouse.CursorPos;
  tmpChild := CreateViewOf2chRes(Text, View.ThreadItem, Point);
  if Assigned(tmpChild) then begin  //tmpChild=nil̏ꍇAviewς݂ŃANZXAVɂȂ\
    if Assigned(View.Child) then
      tmpChild.Free
    else
      View.Child := tmpChild;
    View.Child.OnMouseLeave := View.ChildMouseLeave;
  end;
end;


//̃|bvAbvr[쐻
function TNestingHint.CreateNestPopUpView: TNestPopUpView;
var
  i: Integer;
  c: Integer;
begin
  Result := TNestPopUpView.Create(Owner);
  Result.Parent := Parent;
  Result.PopupMenu := PopUpMenu;
  Result.SetFont(Font.Name, Font.Size);
  Result.ColorEnable := ColorOnFix;
  Result.ColorDisable := Color;
  Result.Color := Color;
  Result.TopMargin := 2;
  Result.LeftMargin := 2;
  Result.RightMargin := 2;
  for i := 0 to 31 do begin
    Result.TextAttrib[i].color := Font.Color;
    Result.TextAttrib[i].style := Font.Style; //qgbold͎gȂ
  end;
  Result.TextAttrib[2].color := LinkColor;
  Result.TextAttrib[2].style := [fsUnderline];
  Result.TextAttrib[3].color := LinkColor;
  Result.TextAttrib[3].style := [fsUnderline];
  c := (Font.color and $ff0000) or
       ((((ColorToRGB(Font.Color) and $ff) * 2 + (ColorToRGB(Color) and $ff) + (ColorToRGB(Color) and $ff)) div 4) and $ff) or
       ((((ColorToRGB(Font.Color) and $ff00) * 2 + (ColorToRGB(Color) and $ff00) + (ColorToRGB(ColorOnFix) and $ff00)) div 4) and $ff00) or
       ((((ColorToRGB(Font.Color) and $ff0000) * 2 + (ColorToRGB(Color) and $ff0000) + (ColorToRGB(ColorOnFix) and $ff0000)) div 4) and $ff0000);
  Result.TextAttrib[26].color := c;
  Result.TextAttrib[26].style := Font.Style;
  Result.TextAttrib[27].color := c;
  Result.TextAttrib[27].style := Font.Style;
  Result.OnAbort := AbortWork;
  Result.OnStatusTextChange := TextViewStatusTextChange;
  Result.OnMouseDown := TextViewMouseDown;
  Result.SetPhysicalCaret(0, 0);
end;


//URLŎw肵X̃|bvAbvr[쐻
function TNestingHint.CreateViewOf2chRes(URI: String; Base: TThreadItem; Point: TPoint): TNestPopUpView;
  procedure ShowBoardInfo(const board: TBoard);
  begin
    AbortWork(nil);
    Result := CreateNestPopUpView;
    Result.Append(URI + #13#10 + TCategory(board.category).name + ' [' + board.name + ']');
  end;
  procedure ShowThreadInfo(const rangearray: TRangeArray; const thread: TThreadItem);
  var
    D2View: TDat2View;
    i: integer;
  begin
    Result := CreateNestPopUpView;
    Result.ThreadItem := thread;
    Result.StartIndex := rangearray[0].st;
    D2View := TDat2View.Create(Result);
    D2View.DD_OFFSET_LEFT := 16;
    D2View.DD_OFFSET_LEFT := 8;

    if Assigned(WorkingD2View) then
      WorkingD2View.Cancel;
    WorkingD2View := D2View;

    D2View.OutPutPopupSpace := Config.ojbOutPutPopupSpace; //184 xG΍
    if (thread <> BaseThread) or (rangearray[0].st = -1) then
      D2View.WriteText(MainWnd.ThreadToURL(thread) + #13#10 +
           TCategory(TBoard(thread.board).category).name + ' ['
         + TBoard(thread.board).name + ']@g'
         + HTML2String(thread.title) + 'h' + #13#10#13#10);
    if Config.hintForOtherThread or (thread = BaseThread) then
      for i := 0 to Length(rangearray) - 1 do
        PickUpRes(D2View, Result.ThreadItem, rangearray[i].st, rangearray[i].ed);
    D2View.Flush;
    {$IFNDEF IE}
    Application.ProcessMessages;//tbJ[΍BReleaseHandleƂł킩
    {$ENDIF}
    D2View.Free;

    if D2View <> WorkingD2View then
      FreeAndNil(Result);

    WorkingD2View := nil;

  end;
var
  Board: TBoard;
  Thread: TThreadItem;
  rangearray: TRangeArray;
begin

  Result := nil;

  if Decode2chURL(URI, Board, Thread, rangearray) then begin
    if Board <> nil then begin
      if Thread <> nil then
        ShowThreadInfo(rangearray, thread)
      else
        ShowBoardInfo(board);
      Board.Release;
    end else
      ShowThreadInfo(rangearray, Base);
  end;

  if Assigned(Result) then begin
    Result.FURL := URI;
    if Result.Trim then
      PopUpViewAt(Result, Point)
    else
      FreeAndNil(Result);
  end;

end;

//c[r[쐻
function TNestingHint.CreateTreeView(AThread: TThreadItem; Index: Integer;
                      OutLine: Boolean; Point: TPoint): TNestPopUpView;
var
  IndexTree: TIndexTree;
  Mask: TBits;
  DatOut: TDat2View;
  Count: Integer;
begin

  if (Index < 1) or (Index > AThread.lines) then begin
    Result := nil;
    Exit;
  end;

  Result := CreateNestPopUpView;
  Result.ThreadItem := AThread;

  IndexTree := TIndexTree.Create;
  IndexTree.AllowDup := Config.ojvAllowTreeDup;
  IndexTree.HeadLineLength := Config.ojvLenofOutLineRes;
  IndexTree.Build(Result.ThreadItem, Index);
  Mask := TBits.Create;
  Mask.Size := Result.ThreadItem.lines + 1;
  IndexTree.Mask := Mask;

  DatOut := TDat2View.Create(Result);
  DatOut.DD_OFFSET_LEFT := 16;
  DatOut.DD_OFFSET_LEFT := 8;

  if OutLine then begin
    Count := IndexTree.OutLine(DatOut, Index);
  end else begin
    IndexTree.ShowRoot := False;
    IndexTree.LevelLimit := 1;
    Count := IndexTree.TreeDatOut(DatOut, Index);
  end;

  DatOut.Flush;
  DatOut.Free;
  IndexTree.Free;
  Mask.Free;

  if (Count > 1) or ((Count > 0) and not(OutLine))  then begin
    if Result.Trim then
      PopUpViewAt(Result, Point)
    else
      FreeAndNil(Result);
  end;
end;


//r[w肵|Cgɑ΂ēK؂Ȉʒuɕ\
procedure TNestingHint.PopUpViewAt(View: TNestPopUpView; Point: TPoint);
var
  Rect: TRect;
begin
  Rect := View.CalcAdjstableSize(Screen.Height, Screen.Width);
  AdjustToTextViewLine(Point, Rect);
  Dec(Point.X, 16);
  if Rect.Right + Point.X > Screen.Width then
    Point.X := Screen.Width - Rect.Right;
  if Rect.Top + Point.Y < 0 then
    Point.Y := -Rect.Top;
  OffsetRect(Rect, Point.X, Point.Y);
  View.PopUp(Rect);
end;


//eLXgqg̕񐶐{\
function TNestingHint.Show2chInfo(URI: String; Base: TThreadItem; Point: TPoint): Boolean;
  procedure ShowBoardInfo(const board: TBoard);
  begin
    ShowTextHint(URI + #13#10 +
             TCategory(board.category).name + ' [' + board.name + ']', Point);
  end;
  procedure ShowThreadInfo(const rangearray: TRangeArray; const thread: TThreadItem);
  var
    s: string;
    i: integer;
  begin
    if Config.hintForOtherThread or (Base = thread) then
    begin
      s := '';
      for i := 0 to Length(rangearray) - 1 do
        s := s + PickUpRes(thread, rangearray[i].st, rangearray[i].ed);
    end;

    if Base <> thread then
      s := MainWnd.ThreadToURL(thread) + #13#10 +
           TCategory(TBoard(thread.board).category).name + ' ['
           + TBoard(thread.board).name + ']@g'
           + HTML2String(thread.title) + 'h' + #13#10
           + #13#10 + s;

    ShowTextHint(s, Point);
  end;
var
  board: TBoard;
  thread: TThreadItem;
  rangearray: TRangeArray;
begin
  if Decode2chURL(URI, Board, Thread, rangearray) then begin
    Result := True;
    if Board <> nil then begin
      if Thread <> nil then
        ShowThreadInfo(rangearray, thread)
      else
        ShowBoardInfo(board);
      Board.Release;
    end else
      ShowThreadInfo(rangearray, Base);
  end else begin
    Result := False;
  end;
end;



//eLXgqgw肵|Cgɑ΂ēK؂Ȉʒuɕ\
procedure TNestingHint.ShowTextHint(const str: string; Point: TPoint; MaxWidth: integer = 0; MaxHeight: integer = 0);
var
  Rect: TRect;
  i: integer;
  s: string;
begin
  s := str;
  i := length(s);
  while 0 < i do
  begin
    case s[i] of
    #13,#10:;
    else
      break;
    end;
    Dec(i);
  end;
  SetLength(s, i);
  if MaxWidth <= 0 then
    MaxWidth := Screen.Width;
  rect := CalcHintRect(MaxWidth, s, nil);
  AdjustToTextViewLine(Point, Rect);
  Dec(Point.X, 8);
  Dec(Point.Y, 4); //THintWindowTNestPopUp̐@vZ@␳

  if Rect.Right + Point.X > Screen.Width then
    Point.X := Screen.Width - Rect.Right;
  if Rect.Top + Point.Y < 0 then
    Point.Y := -Rect.Top;
  OffsetRect(Rect, Point.X, Point.Y);
  ActivateHint(rect, s);
end;


//HogeTextView̍sԂT
procedure TNestingHint.AdjustToTextViewLine(var Point: TPoint; Rect: TRect);
var
  VCLWindow: TWinControl;
  LaidView: THogeTextView;
begin
  VCLWindow := FindVCLWindow(Point);
  if (VCLWindow <> nil) and (VCLWindow is THogeTextView) then begin
    LaidView := THogeTextView(VCLWindow);
    Point := LaidView.ScreenToClient(Point);
    Dec(Point.Y, (Point.Y - LaidView.TopMargin + LaidView.Fraction) mod (Abs(LaidView.Font.Height)
                 + LaidView.ExternalLeading) + Rect.Bottom);
    Point := LaidView.ClientToScreen(Point);
  end else begin
    Dec(Point.Y, Rect.Bottom + 8);
  end;
end;

//URLThreadItemƃXԂɕϊ
function TNestingHint.Decode2chURL(Url: String; var Board: TBoard;
                     var ThreadItem: TThreadItem; var rangearray: TRangeArray): Boolean;
var
  host, bbs, datnum: string;
  oldLog: boolean;
  i: Integer;
begin
  Result := False;
  Board := nil;
  ThreadItem := nil;
  if AnsiStartsStr('http://', Url) then begin

    Get2chInfo(Url, host, bbs, datnum, rangearray, oldLog);

    board := i2ch.FindBoard(host, bbs);
    if board <> nil then begin
      Board.AddRef;
      ThreadItem := board.Find(datnum);
      Result := True;
    end;
  end
  else if AnsiStartsStr('file://', Url) or AnsiStartsStr('about:', Url)
          or AnsiStartsStr('#', Url) then begin
    for i := Length(Url) downto 1 do
  	begin
  	  if Url[i] = '#' then
      begin
        GetRangeFromText(Copy(Url, i + 1, Length(Url) - i), rangearray);
        if rangearray[0].st > 0 then
          Result := True;
        break;
      end;
    end;
  end;
end;


//͈̓X𕶎񉻂Ď擾
function TNestingHint.PickUpRes(thread: TThreadItem; startLine, endLine: integer): string;
var
  StrDatOut:TStrDatOut;
begin
  StrDatOut := TStrDatOut.Create;
  PickUpRes(StrDatOut, thread, startLine, endLine);
  Result := StrDatOut.Text;
  StrDatOut.Free;
end;


//͈̓XTDatOutɏo
function TNestingHint.PickUpRes(dest: TDatOut; thread: TThreadItem;
                                startLine, endLine: integer): Integer;
var
  temp: integer;
  tmpDat: TThreadData;
begin

  Result := 0;

  if thread = nil then
    exit;

  if startLine <= 0 then
  begin
    if endLine > 0 then
      startLine := endLine
    else
      exit;
  end
  else begin
    if endLine <= 0 then
      endLine := startLine
    else if endLine < startLine then
    begin
      temp := startLine;
      startLine := endLine;
      endLine := temp;
    end;
  end;
  if endLine - startLine >= 20 then
    endLine := startLine + 19;
  if not assigned(thread.dat) then
  begin
    thread.AddRef;
    if assigned(thread.dat) then begin
      tmpDat := thread.DupData;
      Result := Dat2HTML.ToDatOut(dest, thread.dat, startLine, endLine - startLine + 1, thread.ABoneArray);
      tmpDat.Free;
    end;
    thread.Release;
  end else
    Result := Dat2HTML.ToDatOut(dest, thread.dat, startLine, endLine - startLine + 1, thread.ABoneArray);
end;

//X`̒~v
procedure TNestingHint.AbortWork(Sender: TObject);
begin
  if Assigned(WorkingD2View) then begin
    WorkingD2View.Cancel;
    WorkingD2View := nil;
  end;
end;

//w肳ꂽURLݕ\URLƔr
function TNestingHint.CompareURL(ABaseThread: TThreadItem; Url: String): Boolean;
begin
  Result := (ABaseThread = FBaseThread) and (Url = FLastURL);
end;


//ŉw̃|bvAbvr[̃}EX[uCxg
procedure TNestingHint.TextViewMouseLeave(Sender: TObject);
begin
  ReleaseHandle;
end;


//|bvAbvr[̃}EX_ECxg
procedure TNestingHint.TextViewMouseDown(Sender: TObject; Button: TMouseButton;
                                         Shift: TShiftState; X, Y: Integer);
var
  View: TNestPopUpView;
  ActiveView: TViewItem;
begin
  View := TNestPopUpView(Sender);
  ActiveView := MainWnd.GetActiveView;
  case Button of
    mbLeft: begin
      if ssDouble in Shift then begin
        View.Selecting := False;
        View.OnMouseLeave(View);
      end else if Assigned(View.Child) then begin
        View.Selecting := False;
        if View.Child.Enabled then begin
          if View.Child.URL = View.StatusText then begin
            if Assigned(ActiveView) and (ActiveView.thread = View.Child.ThreadItem) then
              MainWnd.GetActiveView.ScrollToAnchor(View.Child.StartIndex - 1, True, True)
            else
              MainWnd.NavigateIntoView(View.StatusText, gtOther);
            ReleaseHandle;
          end else begin
            View.Child.Enabled := False;
            TextViewStatusTextChange(Sender, View.StatusText);
          end;
        end else begin
          View.Child.Enabled := True;
        end;
      end else if StartWith('http:', View.StatusText, 1) then begin
        if not ImageForm.GetImage(View.StatusText) then
          MainWnd.OpenByBrowser(View.StatusText);
      {$IFDEF IE}
      end else if not(View.StatusText = '') and ((View.StatusText[1] >= '0') and (View.StatusText[1] <= '9')) then begin
        if Assigned(ActiveView) and (ActiveView.thread = View.ThreadItem) then
          MainWnd.GetActiveView.ScrollToAnchor(StrToIntDef(View.StatusText, 0) - 1, True, True)
        else
          MainWnd.NavigateIntoView(MainWnd.ThreadToURL(View.ThreadItem, False,
                                   False, View.StatusText), gtOther);
        ReleaseHandle;
      {$ELSE}
      end else if StartWith('menu:', View.StatusText, 1) then begin
        if Assigned(ActiveView) and (ActiveView.thread = View.ThreadItem) then
          MainWnd.GetActiveView.ScrollToAnchor(StrToIntDef(copy(View.StatusText,
                                         6, High(Integer)), 0) - 1, True, True)
        else
          MainWnd.NavigateIntoView(MainWnd.ThreadToURL(View.ThreadItem, False,
                                   False, copy(View.StatusText, 6, High(Integer))), gtOther);
        ReleaseHandle;
      {$ENDIF}
      end;
    end;
    mbRight: begin
      if Assigned(View.Child) then
        View.ChildMouseLeave(nil);
      inherited ReleaseHandle;
    end;
  end;
end;


end.
